

''' <summary>
''' Contains functions for dealing with dna sequences
''' </summary>
''' <remarks></remarks>
Public Class DNA


    'http://www.promega.com/biomath/calc11.htm


#Region " Member Variables "

    'These constants are used for search types
    Public Const cMatchNormal As String = "Normal"
    Public Const cMatchRev As String = "Reverse"
    Public Const cMatchComp As String = "Complement"
    Public Const cMatchRevComp As String = "Rev-Comp"

    Public Shared cMinMatchLength As Integer = 12 'hardcoded minimum match lengthj

#End Region



    ''' <summary>
    ''' Changes the dna to a complement the complement c -> g, g-c
    ''' </summary>
    ''' <param name="vSEq"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function DNAComplement(ByVal vSeq As String) As String

        Dim Counter1 As Integer
        Dim tSeqOut As String = ""
        Dim tSeqGiven As String = vSEq.ToLower
        Dim tChar As String



        For Counter1 = 0 To vSEq.Length - 1

            tChar = tSeqGiven.Substring(Counter1, 1)

            If tChar = "a" Then
                tSeqOut = tSeqOut & "t"
            ElseIf tChar = "c" Then
                tSeqOut = tSeqOut & "g"
            ElseIf tChar = "t" Then
                tSeqOut = tSeqOut & "a"
            ElseIf tChar = "g" Then
                tSeqOut = tSeqOut & "c"
            Else
                Throw New Exception(tChar & " not part of dna")
            End If

        Next

        'returning (making sure they are all upppercase)
        Return tSeqOut.ToUpper


    End Function

    ''' <summary>
    ''' Reverses the DNA sequence
    ''' </summary>
    ''' <param name="vSeq">A DNA sequnece</param>
    ''' <returns>A DNA sequence reversed</returns>
    ''' <remarks></remarks>
    Public Shared Function DNAReverse(ByVal vSeq As String) As String

        'using the old vb string reversing function
        Return StrReverse(vSeq).ToUpper

    End Function

    ''' <summary>
    ''' Reverses and does the complement
    ''' </summary>
    ''' <param name="vSeq"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function DNAComplementReverse(ByVal vSeq) As String

        Return DNAComplement(DNAReverse(vSeq))

    End Function

    ''' <summary>
    ''' Uppercase the sequence, trims, removes spaces the in the sequence.
    ''' </summary>
    ''' <param name="vSeq"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function DNAFix(ByVal vSeq As String) As String

        Dim tSeq As String

        'Making upper case
        tSeq = vSeq.ToUpper

        'Triming the ends and removing any spaces in the sequece.
        'the sequence might be from the internet so it will have spaces
        'put in it for readability
        tSeq = tSeq.Replace(" ", "")

        Return tSeq

    End Function

    ''' <summary>
    ''' Goes through a dna string and removes invalid characters AND spaces.
    ''' </summary>
    ''' <param name="vSeq">The DNA sequence</param>
    ''' <returns>The fixed DNA sequence</returns>
    ''' <remarks>Used mainly with data entry w/ copy pastes to clean up any issues</remarks>
    Public Shared Function DNASuperFix(ByVal vSeq As String) As String

        Dim tSeq As String
        Dim SB As New System.Text.StringBuilder

        'Fixing up the string some to make the search shorter
        tSeq = vSeq.ToUpper.Replace(" ", "") 'flipping to upper to make matching easier and uniform


        'This could be done any number of ways. This is one
        For CharCounter As Integer = 0 To tSeq.Length - 1
            'Marching through the string and checking each character

            Dim tChar As String = tSeq.Substring(CharCounter, 1)

            If tChar = "A" Or tChar = "C" Or tChar = "G" Or tChar = "T" Then
                SB.Append(tChar)
            End If

        Next

        'returning
        Return SB.ToString


    End Function

    ''' <summary>
    ''' Determines if all the characters are a valid dna char
    ''' </summary>
    ''' <param name="vSeq">A DNA sequence</param>
    ''' <returns>True if the DNA is valid</returns>
    ''' <remarks></remarks>
    Public Shared Function DNAIsValid(ByVal vSeq As String) As Boolean


        Dim tSeq As String = vSeq.ToUpper.Trim
        Dim tChar As String
        Dim Counter1 As Integer


        DNAIsValid = False 'initially setting to false for early leave potential


        'added 5-27-07 
        'If vSeq = "" Then Exit Function

        For Counter1 = 0 To tSeq.Length - 1

            tChar = tSeq.Substring(Counter1, 1)

            If tChar <> "A" And tChar <> "C" And tChar <> "G" And tChar <> "T" Then
                'Not a valid character for dna
                Exit Function
            End If


        Next Counter1


        'Made it this far
        Return True


    End Function

    ''' <summary>
    ''' Creates a collection of DNA strands for of a particular length to test 
    ''' </summary>
    ''' <param name="vSeq">A DNA sequence</param>
    ''' <returns>A collection full of dna strands</returns>
    ''' <remarks>The idea here is to split the string into all the potential shortest strands for comparison</remarks>
    Public Shared Function DNASplitter(ByVal vSeq As String) As Collection

        'this can get really big because we are chopping a string into the smallest possible match. 
        'think every sequence of minimum lenght starting at 0 and moving ownard
        Dim tCollect As New Collection

        If vSeq.Length <= DNA.cMinMatchLength Then
            'The minimum strand length is > than the dna sequence
            tCollect.Add(vSeq)

            Return tCollect

        Else
            Dim Counter1 As Integer = 0

            For Counter1 = 0 To vSeq.Length - DNA.cMinMatchLength
                'loading up the collection with minimum sequence after minimum sequence.
                'memory warning.
                tCollect.Add(vSeq.Substring(Counter1, DNA.cMinMatchLength))
            Next Counter1


            Return tCollect

        End If


    End Function

    ''' <summary>
    ''' Searches DNA strand for matches
    ''' </summary>
    ''' <param name="vDNARef">is a dna reference</param>
    ''' <param name="vDNASearch">this is the oligo's sequence</param>
    ''' <param name="vMatchType">Rev, normal, rev compl, comp.</param>
    ''' <returns>Generic of dna search results</returns>
    ''' <remarks></remarks>
    Public Shared Function DNASearcher(ByVal vDNARef As String, ByVal vDNASearch As String, ByVal vMatchType As String) As Generic.List(Of DNASearchResult)


        Dim tStr As String 'temporary string to use
        Dim Counter1 As Integer = 0
        Dim CurrentLength As Integer = cMinMatchLength
        Dim Found As Boolean = False 'flag for when 
        Dim tCollect As New Generic.List(Of DNASearchResult)
        Dim tMatchCount As Integer
        Dim EndPosCollect As Collection 'holds all the end positions
        Dim tSR As DNASearchResult 'temporary search result record
        Dim Abort As Boolean 'flag for leaving a search while loop


        If vDNASearch.Length < DNA.cMinMatchLength Then
            'The minimum match is less than the DNAsearch length
            Return tCollect
        End If

        Do While Abort = False

            If Counter1 + cMinMatchLength > vDNASearch.Length Then
                'Reached the end of the string nothing more to search on
                Exit Do
            End If

            tStr = vDNASearch.Substring(Counter1, cMinMatchLength)
            CurrentLength = cMinMatchLength


            If vDNARef.IndexOf(tStr) <> -1 Then
                tMatchCount = DNA.MatchCount(vDNARef, tStr)

                Found = True


                'try to go one bigger
                Do While Found = True

                    If Counter1 + CurrentLength + 1 > vDNASearch.Length Then
                        'Reached the end of the search string
                        Exit Do
                    End If

                    'Adding one character to the searchstring
                    tStr = vDNASearch.Substring(Counter1, CurrentLength + 1)

                    If vDNARef.IndexOf(tStr) <> -1 Then
                        'Matched on a 1 character longer string
                        CurrentLength += 1



                        'tMatchCount = tCurrentMatchCount

                    Else
                        'Didn't match after adding one character
                        Found = False




                        Counter1 = Counter1 + 1 '+ CurrentLength - cMinMatchLength

                    End If

                Loop



                'Found the longest, now need to back track
                For BackCounter As Integer = tStr.Length To DNA.cMinMatchLength Step -1

                    tStr = tStr.Substring(0, BackCounter)


                    If vDNARef.IndexOf(tStr) <> -1 Then


                        'Loading up the record
                        tSR = New DNASearchResult(vMatchType, tStr, tMatchCount, "")


                        'Creating the end positions
                        EndPosCollect = New Collection


                        For Each tEndPos As Integer In DNA.EndPositions(vDNARef, tStr)
                            If tEndPos = 31 Then
                                Dim asdf As Boolean = True
                            End If
                            If DNASearchResult.AllFindEndposition(tCollect, tEndPos, tStr) = False Then
                                EndPosCollect.Add(tEndPos)
                            End If
                        Next


                        If EndPosCollect.Count > 0 Then
                            tSR.EndPositions = EndPosCollect
                            tCollect.Add(tSR)
                        End If


                    End If


                Next

                Counter1 += 1

            Else
                'Didn't have a mtach on the string
                Counter1 += 1

            End If 'vDNARef.IndexOf(tStr) <> -1

        Loop


        Return tCollect

    End Function

    ''' <summary>
    ''' Does the dna search for normal, revers, compliment, reverse compliment
    ''' </summary>
    ''' <param name="vDNARef">is a reference dna sequence</param>
    ''' <param name="vDNASearch">Is the oligos sequence</param>
    ''' <returns>A generic list of dna search results</returns>
    ''' <remarks></remarks>
    Public Shared Function DNACompleteSearch(ByVal vDNARef As String, ByVal vDNASearch As String) As Generic.List(Of DNASearchResult)

        Dim tOut As New Generic.List(Of DNASearchResult)


        For Each t1 As DNASearchResult In DNA.DNASearcher(vDNARef, vDNASearch, DNA.cMatchNormal)
            tOut.Add(t1)
        Next


        For Each t2 As DNASearchResult In DNA.DNASearcher(vDNARef, DNA.DNAReverse(vDNASearch), DNA.cMatchRev)
            tOut.Add(t2)
        Next

        For Each t3 As DNASearchResult In DNA.DNASearcher(vDNARef, DNA.DNAComplement(vDNASearch), DNA.cMatchComp)
            tOut.Add(t3)
        Next

        For Each t4 As DNASearchResult In DNA.DNASearcher(vDNARef, DNA.DNAComplementReverse(vDNASearch), DNA.cMatchRevComp)
            tOut.Add(t4)
        Next


        Return tOut

    End Function




    ''' <summary>
    ''' Counts the number of times a specific string is in a string. 
    ''' </summary>
    ''' <param name="vLookIn"></param>
    ''' <param name="vSearchStr"></param>
    ''' <param name="vFirstPos"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function MatchCount(ByVal vLookIn As String, ByVal vSearchStr As String, Optional ByVal vFirstPos As Integer = -1) As Integer

        'Note: There is probably a better way to do this but i couldn't find it 
        '      easily

        Dim LastPos As Integer
        Dim tCount As Integer = 0
        Dim Found As Boolean = False

        If vFirstPos <> -1 Then
            LastPos = vFirstPos
            tCount = 1
        Else
            LastPos = 0
        End If


        LastPos = vLookIn.IndexOf(vSearchStr, LastPos)

        Do While LastPos <> -1
            LastPos = LastPos + vSearchStr.Length
            tCount += 1
            LastPos = vLookIn.IndexOf(vSearchStr, LastPos)
        Loop

        Return tCount

    End Function




    Public Shared Function EndPositions(ByVal vLookIn As String, ByVal vSearchStr As String, Optional ByVal vFirstPos As Integer = -1) As Collection


        'Note: There is probably a better way to do this but i couldn't find it 
        '      easily

        Dim LastPos As Integer
        Dim tCount As Integer = 0
        Dim Found As Boolean = False
        Dim tCollect As New Collection


        If vFirstPos <> -1 Then
            LastPos = vFirstPos
            'tCount = 1
            tCollect.Add(LastPos + vSearchStr.Length)
            'tCollect.Add(LastPos)
        Else
            LastPos = 0
        End If


        LastPos = vLookIn.IndexOf(vSearchStr, LastPos)

        Do While LastPos <> -1
            'Zerobased here
            tCollect.Add(LastPos + vSearchStr.Length - 1)
            'tCollect.Add(LastPos)
            LastPos = LastPos + vSearchStr.Length

            'tCount += 1
            LastPos = vLookIn.IndexOf(vSearchStr, LastPos)
        Loop

        Return tCollect

    End Function



#Region " Temperature Stuff "

    ''' <summary>
    ''' Counts just the number of C's and G's in a sequence
    ''' </summary>
    ''' <param name="vSeq"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function CountGandCs(ByVal vSeq As String) As Integer

        Dim tSeq As String = DNAFix(vSeq)


        For Counter1 As Integer = 0 To tSeq.Length - 1
            Dim tChar As String = vSeq.Substring(Counter1, 1)

            If tChar = "G" Or tChar = "C" Then
                CountGandCs += 1
            End If

        Next

    End Function


    ''' <summary>
    ''' Counts just eh number of t's and a's in a dna sequence
    ''' </summary>
    ''' <param name="vSeq"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function CountTandAs(ByVal vSeq As String) As Integer

        Dim tSeq As String = DNAFix(vSeq)


        For Counter1 As Integer = 0 To tSeq.Length - 1
            Dim tChar As String = vSeq.Substring(Counter1, 1)

            If tChar = "T" Or tChar = "A" Then
                CountTandAs += 1
            End If

        Next

    End Function

    ''' <summary>
    ''' The basic temperature formula
    ''' </summary>
    ''' <param name="vSeq"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function BasicTm(ByVal vSeq As String) As Single

        If vSeq.Length < 14 Then

            'Tm = 4C  x  (number of Gs and Cs in the primer) + 2C  x  (number of As and Ts in the primer)
            Return 4 * (DNA.CountGandCs(vSeq) + 2 * DNA.CountTandAs(vSeq))

        Else
            'Tm =  64.9C + 41C x (number of Gs and Cs in the primer  16.4)/N

            Return 64.9 + 41 * (DNA.CountGandCs(vSeq) - 16.4) / vSeq.Length

        End If


    End Function

    ''' <summary>
    ''' This is the new melting point method from Dr. Kane
    ''' </summary>
    ''' <param name="vSeq"></param>
    ''' <param name="MofNaCl"></param>
    ''' <param name="vFormamide"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function GoodTm(ByVal vSeq As String, ByVal MofNaCl As Single, ByVal vFormamide As Single) As Single


        'Calc Tm (in Celsius) = 81.5 + 16.6log(M of NaCl) + 0.41(%GC) - 0.65(%formamide) - 500/oligolength

        'Terms in the calculation;

        'M of NaCl: I use 0.375 M (molarity), but this can be almost anything - FIND THE Na+ or NaCl for standard PCR

        'GC% is number of Gs and Cs divided by total character count (GTCGTAGAATAGTC = 6/14(*100) = 42.9

        'Formamide is sometimes added to DNA hybridization reactions, but this term can be omitted if needed (default is zero - so the term is zero)


        Return 81.5 + 16.6 * Math.Log(MofNaCl) + (0.41 * DNA.CountGandCs(vSeq) / vSeq.Length * 100) - 0.65 * (vFormamide) - (500 / vSeq.Length)


    End Function



#End Region



    Public Shared Function PCRSearch(ByVal vDNARef As String, ByVal vDNASearch As String, ByVal vMatchType As String, ByVal vMinMatchLength As String) As Generic.List(Of DNASearchResult)

        Dim Abort As Boolean = False
        Dim MString As String
        Dim CLength As Integer = vDNASearch.Length
        Dim MPoint As Integer
        Dim tCollect As New Generic.List(Of DNASearchResult)
        Dim RevString As String = StrReverse(vDNASearch)
        Dim tEndPos As Integer


        Do While CLength > vMinMatchLength

            'Note: Going in reverse because have to bind on the 3' end.
            '      shorter matches will be blocked from the collection
            MString = vDNASearch.Substring(vDNASearch.Length - CLength)



            MPoint = vDNARef.IndexOf(MString)

            If MPoint <> -1 Then
                'Found at least one match

                tEndPos = MPoint + MString.Length
                'Checking if the start position is already in the collect
                If DNASearchResult.FindStopPos(tCollect, tEndPos) Is Nothing Then
                    tCollect.Add(New DNASearchResult(vMatchType, MString, 1, MPoint))
                    tCollect.Item(tCollect.Count - 1).StartPosition = MPoint
                    tCollect.Item(tCollect.Count - 1).StopPosition = tEndPos
                End If

                'Try to find the match somewhere else in the dnaref
                MPoint = vDNARef.IndexOf(MString, MPoint + 1)

                Do While MPoint <> -1
                    'found it somewhere else

                    tEndPos = MPoint + MString.Length

                    If DNASearchResult.FindStopPos(tCollect, tEndPos) Is Nothing Then
                        tCollect.Add(New DNASearchResult(vMatchType, MString, 1, MPoint))
                        tCollect.Item(tCollect.Count - 1).StartPosition = MPoint
                        tCollect.Item(tCollect.Count - 1).StopPosition = tEndPos
                    End If

                    MPoint = vDNARef.IndexOf(MString, MPoint + 1)

                Loop

            End If

            CLength -= 1

        Loop


        Return tCollect


    End Function


End Class


''' <summary>
''' Is a dna search result.
''' </summary>
''' <remarks></remarks>
Public Class DNASearchResult

    Private m_MatchType As String = ""
    Private m_MatchString As String = ""
    Private m_MatchCount As Integer
    Private m_MatchIndexes As String = ""
    'Private m_StartPos As Integer = 0

    Public EndPositions As Collection

    Public MatchLongestIndex As String = ""

    Public StartPosition As Integer 'This is just for PCR stuff
    Public StopPosition As Integer 'This is for pcr

    Public Sub New(ByVal vMatchType As String, ByVal vMatchString As String, ByVal vMatchCount As Integer, ByVal vMatchIndexes As String)

        Me.m_MatchString = vMatchString
        Me.m_MatchType = vMatchType
        Me.m_MatchCount = vMatchCount
        Me.m_MatchIndexes = vMatchIndexes

        'Me.m_StartPos = vStartPos


    End Sub

#Region " Properties "

    Public ReadOnly Property MatchType() As String
        Get
            Return Me.m_MatchType
        End Get
    End Property

    Public ReadOnly Property MatchLength() As Integer
        Get
            Return MatchString.Length
        End Get
    End Property

    Public Property MatchString() As String
        Get
            Return Me.m_MatchString
        End Get
        Set(ByVal value As String)
            Me.m_MatchString = value
        End Set
    End Property

    Public ReadOnly Property Blurb() As String
        Get
            Return Me.MatchString & "-" & Me.MatchType
        End Get
    End Property

    Public ReadOnly Property MatchCount() As Integer
        Get
            Return Me.EndPositions.Count
        End Get

    End Property

    Public ReadOnly Property MatchIndexes() As String
        Get
            Dim tStr As String = ""

            For Each tEndPos As Integer In Me.EndPositions
                tStr = tStr & "(" & (tEndPos + 1) - Me.MatchLength + 1 & "-" & tEndPos + 1 & "), "
            Next

            Return tStr.Substring(0, tStr.Length - 2)

        End Get

    End Property

    Public Shared Function LongestMatchPositions(ByRef rResults As Generic.List(Of DNASearchResult)) As String


        Dim LongestRec As DNASearchResult
        Dim LongestMatch As Integer

        For Each tR As DNASearchResult In rResults
            If tR.MatchLength > LongestMatch Then
                LongestRec = tR
            End If

        Next

        Return LongestRec.MatchIndexes


    End Function



#End Region



    ''' <summary>
    ''' Determines the highest match value
    ''' </summary>
    ''' <param name="rResults"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function GetLongest(ByRef rResults As Generic.List(Of DNASearchResult)) As Integer

        Dim tLongest As Integer

        For Each tR As DNASearchResult In rResults
            If tR.MatchLength > tLongest Then
                tLongest = tR.MatchLength
            End If
        Next

        Return tLongest

    End Function

    ''' <summary>
    ''' Finds a particular end position
    ''' </summary>
    ''' <param name="vEndPos"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function FindEndPosition(ByVal vEndPos As String) As Boolean

        For Each tS As Integer In Me.EndPositions
            If tS = vEndPos Then
                Return True
            End If
        Next

        Return False

    End Function


    Public Shared Function AllFindEndposition(ByRef rSearchRecs As Generic.List(Of DNASearchResult), ByVal vEndPos As Integer, ByVal vMatchStr As String) As Boolean

        For Each tR As DNASearchResult In rSearchRecs

            For Each tEndPos As Integer In tR.EndPositions
                If vEndPos >= tEndPos - tR.MatchLength And vEndPos <= tEndPos Then
                    Return True
                End If

            Next

            'If tR.FindEndPosition(vEndPos) = True or Then
            'Return True
            'End If
            'End If
        Next

        Return False


    End Function



    Public Shared Function FindStartPos(ByRef rCollect As Generic.List(Of DNASearchResult), ByVal vStartPosition As Integer) As DNASearchResult

        For Each tR As DNASearchResult In rCollect

            If tR.StartPosition = vStartPosition Then
                Return tR
            End If
        Next

        'Made it this far
        Return Nothing

    End Function

    Public Shared Function FindStopPos(ByRef rCollect As Generic.List(Of DNASearchResult), ByVal vStopPosition As Integer) As DNASearchResult

        For Each tR As DNASearchResult In rCollect

            If tR.StopPosition = vStopPosition Then
                Return tR
            End If
        Next

        'Made it this far
        Return Nothing

    End Function



    Public Function DistanceBetween(ByRef rSR As DNASearchResult) As Integer

        'return rsr.StartPosition - me.StopPosition

        If rSR.StartPosition > Me.StopPosition Then
            'rsr is after this
            Return rSR.StartPosition - Me.StopPosition
        ElseIf rSR.StopPosition < Me.StartPosition Then
            'rsr is before this
            Return rSR.StopPosition - Me.StartPosition
        End If

    End Function




End Class



Public Class PrimerPair


    Public Forward As DNASearchResult
    Public Backward As DNASearchResult


    Public Sub New(ByRef rForward As DNASearchResult, ByRef rBackward As DNASearchResult)

        Forward = rForward
        backward = rBackward

    End Sub


End Class



